#!/usr/bin/python
"""
    Basic (POC) OPC UA shape detection sensor using opencv libraryś countours detection.
    Exposes a default OPC UA shape detection profile where:
    0.0 - nothing detected
    1.0 - triangle
    2.0 - rectangle
    3.0 - circle

    Based on this classification any PLC can integrate this Optical Inspection and implement
    additional logic.
"""

import cv2
import numpy as np
import asyncio
import logging
from asyncua import Server

def nothing(x):
    # any operation
    pass

async def main():
    _logger = logging.getLogger(__name__)
    # setup our server
    server = Server()
    await server.init()
    server.set_endpoint("opc.tcp://0.0.0.0:4840/freeopcua/server/")

    # set up our own namespace, not really necessary but should as spec
    uri = "http://examples.freeopcua.github.io"
    idx = await server.register_namespace(uri)

    # populating our address space
    myobj = await server.nodes.objects.add_object(idx, "OpticalInspection")
    myvar = await myobj.add_variable(idx, "shape", 0.0)
    await myvar.set_writable()

    # init camera
    cap = cv2.VideoCapture(0)
    cv2.namedWindow("Trackbars")
    cv2.createTrackbar("L-H", "Trackbars", 0, 180, nothing)
    cv2.createTrackbar("L-S", "Trackbars", 0, 255, nothing)
    cv2.createTrackbar("L-V", "Trackbars", 3, 255, nothing)
    cv2.createTrackbar("U-H", "Trackbars", 39, 180, nothing)
    cv2.createTrackbar("U-S", "Trackbars", 155, 255, nothing)
    cv2.createTrackbar("U-V", "Trackbars", 148, 255, nothing)
    font = cv2.FONT_HERSHEY_COMPLEX

    _logger.info("Starting server!")
    async with server:

        while True:
            # XXX: find out why we need to sleep (otherwise OPC UA server stops work)
            await asyncio.sleep(0.0001)

            # read and process camera 
            _, frame = cap.read()
            hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
            l_h = cv2.getTrackbarPos("L-H", "Trackbars")
            l_s = cv2.getTrackbarPos("L-S", "Trackbars")
            l_v = cv2.getTrackbarPos("L-V", "Trackbars")
            u_h = cv2.getTrackbarPos("U-H", "Trackbars")
            u_s = cv2.getTrackbarPos("U-S", "Trackbars")
            u_v = cv2.getTrackbarPos("U-V", "Trackbars")
            lower_red = np.array([l_h, l_s, l_v])
            upper_red = np.array([u_h, u_s, u_v])

            mask = cv2.inRange(hsv, lower_red, upper_red)
            kernel = np.ones((5, 5), np.uint8)
            mask = cv2.erode(mask, kernel)

            # Contours detection
            contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

            for cnt in contours:
                area = cv2.contourArea(cnt)
                approx = cv2.approxPolyDP(cnt, 0.02*cv2.arcLength(cnt, True), True)
                x = approx.ravel()[0]
                y = approx.ravel()[1]

                if area > 400:
                    cv2.drawContours(frame, [approx], 0, (0, 0, 0), 5)
                    if len(approx) == 3:
                        cv2.putText(frame, "Triangle", (x, y), font, 1, (0, 0, 0))
                        await myvar.write_value(1.0)
                        #_logger.info("Triangle")
                    elif len(approx) == 4:
                        cv2.putText(frame, "Rectangle", (x, y), font, 1, (0, 0, 0))
                        await myvar.write_value(2.0)
                        #_logger.info("Rectangle")
                    elif 7 < len(approx) < 20:
                        cv2.putText(frame, "Circle", (x, y), font, 1, (0, 0, 0))
                        await myvar.write_value(3.0)
                        #_logger.info("Circle")
                    else:
                        await myvar.write_value(0.0)

            cv2.imshow("Frame", frame)
            cv2.imshow("Mask", mask)

            key = cv2.waitKey(1)
            if key == 27:
                break

if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    asyncio.run(main(), debug=True)
