#!//usr/bin/python
"""
Basic OPC UA <-> HTTP gateway server.
"""

import sys
import asyncio
import asyncua
from dataclasses import dataclass, field
import json
import requests
import urllib
import argparse
import logging
import __main__

# command line handling
parser = argparse.ArgumentParser(description='Run OPCUA Server.')
a = parser.add_argument
a('--ipv4', help='The IPv4 address on which the OPCUA Server runs', default="0.0.0.0")
a('--ipv6', help='The IPv6 address on which the OPCUA Server runs', default="::")
a('--ipv6-enabled', help='The IPv6 address check whether it is enabled or disabled', default="1")
a('--port', help='The port on which the OPCUA Server runs', default="4840")
a('--xml', help='Path of XML to configure Server. See asyncua doc for more details.', default=None)
a('--erp5-url', help='URL of ERP5 instance to which data shall be send.', default=None)
a('--erp5-username', help='Username of ERP5 instance to which data shall be send.', default=None)
a('--erp5-password', help='Password of ERP5 instance to which data shall be send.', default=None)
args = parser.parse_args()
ipv4 = args.ipv4
ipv6 = args.ipv6
ipv6_enabled = args.ipv6_enabled
port = args.port
xml = args.xml
erp5_url = args.erp5_url
erp5_username = args.erp5_username
erp5_password = args.erp5_password
ERP5_REQUEST_API = "ERP5Site_handleOPCUARequest"

# ERP5 backend storage for OPCUA Document
@dataclass(frozen=True)
class ERP5Handler(asyncua.common.subscription.SubHandler):
  url: str
  session: requests.Session = field(default_factory=requests.Session)

  @property
  def uri(self):
    if self.url is not None:
      return f"%s/%s" %(self.url, ERP5_REQUEST_API)

  def call(self, http_method="POST", **data):
    serialized_data = {}

    for k, v in data.items():
      if not isinstance(v, (int, float, str)):
        v = str(v)
      
      serialized_data[k] = v
      print(f"Key: {k}\nValue: {v}\nValue Type: {type(v)}\n\n")

    params = urllib.parse.quote_plus(json.dumps(serialized_data))
    self.session.auth = (erp5_username, erp5_password)

    if http_method == "POST":
        self.session.post(f"{self.uri}?data={params}")
    elif http_method == "GET":
        return self.session.get(f"{self.uri}")

  def datachange_notification(self, node, val, data):
    self.call(node=node, val=val, data=data)

  def event_notification(self, event):
    self.call(event=event)

# init main ERP5 storage
erp5_handler = ERP5Handler(erp5_url)

class InternalSession(asyncua.server.internal_session.InternalSession):

  async def read(self, params):
    erp5_handler.call(params=params)
    return await super().read(params)

# Start OPCUA Server
async def main():
  _logger = logging.getLogger(__name__)

  # setup our server
  server = asyncua.Server()
  await server.init()

  if bool(int(ipv6_enabled)):
    server.set_endpoint(f"opc.tcp://[{ipv6}]:{port}/freeopcua/server/")
  else:
    server.set_endpoint(f"opc.tcp://{ipv4}:{port}/freeopcua/server/")

  if xml is not None:
    await server.import_xml(xml)

  # read previous state as saved in ERP5 backend
  erp5_json = erp5_handler.call(http_method="GET").json()
  _logger.error(erp5_json)
  for k,v in erp5_json.items():
    # set
    node = server.get_node(k)
    _logger.debug("Init from ERP5. Set %s = %s at %s" %(k, v, node))
    await node.write_value(v)

  subscription = await server.create_subscription(1000, erp5_handler)
  await subscription.subscribe_events()
  nodes = await asyncua.common.ua_utils.get_nodes_of_namespace(server)
  await subscription.subscribe_data_change(nodes)

  def create_session(name,
                    user=asyncua.server.users.User(role=asyncua.server.users.UserRole.Anonymous),
                    external=False):
    self = server.iserver
    return InternalSession(self, self.aspace, self.subscription_service, name, user=user, external=external)

  server.iserver.create_session = create_session

  # start OPCUA server
  _logger.info("Starting server!")
  async with server:
    while True:
      await asyncio.sleep(1)

logging.basicConfig(level=logging.DEBUG)
asyncio.run(main(), debug=True)

if __name__ == '__main__':
    sys.exit(__main__.main())
